/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor.sybase;

import org.xquark.extractor.ExtractorReadOnlyConnection;
import org.xquark.extractor.algebra.*;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLResultSet;
import org.xquark.xquery.parser.XQueryExpression;

public class ExtractorStatement extends org.xquark.extractor.ExtractorStatement 
{
	private static final String RCSRevision = "$Revision: 1.6 $";
	private static final String RCSName     = "$Name:  $";

    protected TempTableIDGenerator _tmpTableIDProvider;
    // NOTE: do not set to null since init is performed by super constructor

    /**
	 * Execute the Xquery request and return an XML document as result.
	 *
	 * @param xquery the query to execute.
	 * @return the XML result document.
	 */

	public ExtractorStatement(ExtractorReadOnlyConnection connection, int stmtId)
	throws XMLDBCException {
		super(connection, stmtId);
	}
	
    protected void createVisitors() 
    {
        TypeInterpreter ti = new TypeInterpreter();
        _genAlgebraVisitor = new GenAlgebraVisitor( ti,_metadataManager, _attIDProvider, _relIDProvider);

        _algebraTypeVisitor     = new AlgebraTypeVisitor();
        _removeProjectVisitor   = new RemoveIntermediateProjectVisitor();
        _brav                   = new BuildReferredAttributes();
        _apv                    = new AnalyzePredicateVisitor();

        _tmpTableIDProvider = new TempTableIDGenerator();
        _genSqlVisitor = new org.xquark.extractor.sybase.SybaseGenSqlVisitor(_relIDProvider, _tmpTableIDProvider);
        _genSqlVisitor.setMetadataManager(_metadataManager);
    }
    	
    public XMLResultSet executeXQuery(XQueryExpression xExpr) throws XMLDBCException {
        _tmpTableIDProvider.reset();
        return super.executeXQuery(xExpr);
	}
 }
