/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator.plan;

import java.util.ArrayList;

import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xquark.mediator.MediatorException;
import org.xquark.mediator.DOMUtils.FillExprWithTupleVisitor;
import org.xquark.mediator.DOMUtils.SAX2DOM;
import org.xquark.mediator.DOMUtils.Tuple;
import org.xquark.mediator.decomposer.Utils;
import org.xquark.schema.validation.PSVInfoSetProvider;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.schema.validation.ValidatingSchemaHandler;
import org.xquark.xml.xdbc.*;
import org.xquark.xquery.parser.XQueryExpression;

public class OpSource extends ZeroOp implements Runnable {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.21 $";
	private static final String RCSName = "$Name:  $";

	// **********************************************************************
	// * CLASS VARIABLES
	// **********************************************************************
	private XMLConnection connection = null;
	private XMLStatement statement = null;
	private String source = null;

	// ***********************************************************************
	// * INITIALIZATION
	// ***********************************************************************
	/**
	 *
	 */
	public OpSource(ExecutionPlan plan, XQueryExpression expression, ArrayList origpaths, String source, boolean hasIdentifier) throws MediatorException {
		super(plan, expression);

		addPaths(origpaths);
		/*
		if (origpaths != null)
			for (int i = 0 ; i < origpaths.size() ; i ++) {
				XQueryExpression xi = (XQueryExpression)origpaths.get(i) ;
				if (!pathslist.contains(xi.toString())) {
					paths.add(xi) ;
					pathslist.add(xi.toString());
				}
			}
		 */
		this.source = source;
		this.size = this.getPaths().size();
		//this.idsize = this.getVariables().size();
		//this.idsize = 1;
		if (hasIdentifier)
			this.idsize = 1;
		/*
		ArrayList idvars = plan.getIdVars();
		if (idvars == null || idvars.isEmpty()) this.idsize = 0;
		else {
			ArrayList vars = ((FLWRExpression)expression).getVariables();
			for (int i=0;i<vars.size();i++)
				if (idvars.contains((Variable)vars.get(i))) idsize++;
		}
		 */
		if (source == null)
			throw new MediatorException("AlgSource: no source !");

	}

	// #############################################################################
	// VISITOR STUFF
	// #############################################################################

	public void accept(OperatorVisitor visitor) throws MediatorException {
		visitor.visit(this);
	}

	// ***********************************************************************
	// * EXECUTE QUERY
	// ***********************************************************************
	/**
	 *
	 */
	public void executeQuery() throws MediatorException {
		super.executeQuery();

		//System.out.println ("1.RESULT SET " + resultset) ;
		try {
			if (connection == null) {
				connection = plan.getConnection(source);
				if (connection == null)
					throw new MediatorException("AlgSource: couldn't established connection to " + source);
			}
			// change LARS 05/04/04
			//if (statement == null)
			if (statement != null)
				statement.close();
			statement = connection.createStatement();
			resultset = new SourceResultSet(this, statement);
		} catch (XMLDBCException e) {
			try {
				throw new MediatorException("Can't construct " + getClass() + ": " + e.getMessage() + " for " + source + "(" + connection.getURL() + ")" + " with expression " + expression, e);
			} catch (XMLDBCException xe) {
				throw new MediatorException("Can't construct " + getClass() + ": " + e.getMessage() + " for " + source + "(" + connection + ")" + " with expression " + expression, e);
			}
		}
	}

	public void terminate() {
		try {
			if (statement != null) {
				statement.close(); 
				statement = null;
			}
		} catch (XMLDBCException e) {
		}
//		try {
//			if (connection != null) {
//				connection.close(); 
//				connection = null;
//			}
//		} catch (XMLDBCException e) {
//		}
	}
	// **********************************************************************
	// * OPTIMIZE
	// **********************************************************************
	/**
	 * Return all the sources depending of this node.
	 */
	public ArrayList getSources() {
		if (sources == null) {
			sources = new ArrayList();
			sources.add(source);
		}
		return sources;
	}

	public String toCompleteString(int indent) {
		StringBuffer buf = new StringBuffer();
		buf.append(Utils.makeIndent(indent) + "<" + getClass().getName() + " isLet=\"" + islet + "\" valueDepends=\"" + valueDepends + "\" whereDepends=\"" + whereDepends + "\">\n") ;
		buf.append(Utils.makeIndent(indent + 1) + "<Expression>\n");
		buf.append(Utils.makeIndent(indent + 2) + expression + "\n");
		buf.append(Utils.makeIndent(indent + 1) + "</Expression>\n");
		buf.append(Utils.makeIndent(indent + 1) + "<Source>" + source + "</Source>\n");
//		if (variables != null) {
//			buf.append(Utils.makeIndent(indent + 1) + "<Variables>\n");
//			for (int i = 0; i < variables.size(); i++) {
//				buf.append(Utils.makeIndent(indent + 2) + "<Variable>" + variables.get(i) + "</Variable>\n");
//			}
//			buf.append(Utils.makeIndent(indent + 1) + "</Variables>\n");
//		}
		if (paths != null) {
			buf.append(Utils.makeIndent(indent + 1) + "<Paths>\n");
			for (int i = 0; i < paths.size(); i++) {
				buf.append(Utils.makeIndent(indent + 2) + "<Path>" + paths.get(i) + "</Path>\n");
			}
			buf.append(Utils.makeIndent(indent + 1) + "</Paths>\n");
		}
		buf.append(Utils.makeIndent(indent) + "</" + getClass().getName() + ">\n");
		return buf.toString();
	}

	protected void finalize() {
		try {
			if (null != resultset) {
				resultset.close();
				resultset = null;
			}
		} catch (Exception ex) {};
		try {
			if (null != statement) {
				statement.close();
				statement = null;
			}
		} catch (Exception ex) {}
//		try {
//			if (null != connection) {
//				connection.close();
//				connection = null;
//			}
//		} catch (Exception ex) {}
	}

}

class SourceResultSet extends ZeroResultSet {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.21 $";
	private static final String RCSName = "$Name:  $";

	// **********************************************************************
	// * CLASS VARIABLES
	// **********************************************************************
	private XMLResultSet resultset = null;
	private PSVInfoSetProvider psvip = null;
	//private SchemaValidationContext svc = null;
	private ValidatingSchemaHandler vsh = null;
	private SAX2DOM thandler = null;
	// ***********************************************************************
	// * INITIALIZATION
	// ***********************************************************************
	/**
	 *
	 */
	public SourceResultSet(OpSource algebra, XMLStatement statement) throws XMLDBCException {
		super(algebra);
		
		StringBuffer sb = new StringBuffer();
		String tmpstr = this.operator.getPlan().getDeclarations();
		if (tmpstr != null)
			sb.append(tmpstr);
		sb.append("declare validation preserve;\n");
		// change LARS 06/04/04
		//sb.append("declare namespace ns = \"http://www.xquark.org/EXECUTIONPLAN\";\n");
		sb.append("declare namespace xsi = \"http://www.w3.org/2001/XMLSchema-instance\";\n");
		
		// because in extractor, the variables are renamed, we must
		// tell him not to rename
		if (this.operator.getPlan().getCurrentTuples().isEmpty())
			sb.append(expression.toString());
		else {
			try {
				XQueryExpression tmpExpr = (XQueryExpression) expression.clone();
				ArrayList tmplist = algebra.getPlan().getCurrentTuples();
				/*
				org.xquark.xml.misc.Debug.print("tmplist = " + tmplist);
				for (int i=0;i<tmplist.size();i++)
					org.xquark.xml.misc.Debug.print("tmplist["+i+"] = " + ((Tuple)tmplist.get(i)).toCompleteString());
				*/
				FillExprWithTupleVisitor fewtvisitor = new FillExprWithTupleVisitor(this.operator.getPlan().getSchemaManager(), this.operator.getPlan().getTypeVisitor(), tmplist);
				tmpExpr.accept(fewtvisitor);
				sb.append(tmpExpr.toString());
			} catch (Exception e) {
				throw new XMLDBCException(e.getMessage());
			}
		}
		org.xquark.util.misc.Debug.print("EXECUTING \n" + sb.toString()) ;
		resultset = statement.executeQuery(sb.toString(), XMLConnection.MEDIATOR_STRING_TYPE);
		if (!resultset.hasNext()) {
			XMLStatement stat = resultset.getStatement();
			resultset.close();
			resultset = null;
			stat.close();
			stat = null;
			setFinishedWhenEmpty();
		}
		/*
		if (stat instanceof org.xquark.extractor.ExtractorStatement)
			resultset = stat.executeQuery(sb.toString(), -1) ;
		else
			resultset = stat.executeQuery(sb.toString()) ;
		*/
	}

	// ***********************************************************************
	// * EVALUATE IMPLEMENTATION
	// ***********************************************************************
	/**
	 *
	 * @param non_blocking if the evaluation is blocking, we must
	 *        generate all the results before passing.
	 */
	//int tupleindex = 0;
	protected void evaluate(boolean non_blocking) throws XMLDBCException {
		
		if (resultset == null) return;
		if (!resultset.hasNext()) {
			XMLStatement stat = resultset.getStatement();
			resultset.close();
			resultset = null;
			stat.close();
			stat = null;
			setFinishedWhenEmpty();
			return;
		}
		
		//non_blocking = false;
		
		if (this.operator.isLet())
			non_blocking = false;
		try {
			if (thandler == null) {
				psvip = this.resultset.getPSVInfoSetProvider();
				if (psvip == null) {
					SchemaValidationContext svc = new SchemaValidationContext(this.operator.getPlan().getSchemaManager());
					vsh = new ValidatingSchemaHandler(svc, false);
					thandler = new SAX2DOM(svc);
					vsh.setContentHandler(thandler);
					vsh.setErrorHandler(new ErrorHandler() {
						public void warning(SAXParseException ex) throws SAXException {}
						public void error(SAXParseException ex) throws SAXException {
							System.err.println(ex);
						}
						public void fatalError(SAXParseException ex) throws SAXException {}
					});
					resultset.setContentHandler(vsh);
				} else {
					thandler = new SAX2DOM(psvip);
					resultset.setContentHandler(thandler);
				}
			}
			while (true) {
				Tuple tuple = new Tuple(this);
				//thandler.init(steps, tuple) ;
				thandler.init(tuple);
				if (resultset.hasNext()) {
					if (vsh != null) vsh.startDocument(); else thandler.startDocument();					
					resultset.nextAsSAX();
					if (vsh != null) vsh.endDocument(); else thandler.endDocument(); 
					if (this.operator.isLet()) {
						if (this.buftuples.size() == 0) {
							//tuple.fillIdentifiers(buftuples.getIdentifier());
							tuple.fillIdentifiers();
							buftuples.add(tuple);
						} else {
							Tuple first = (Tuple) this.buftuples.get(0);
							first.appendTuple(tuple);
						}
					} else {
						//tuple.fillIdentifiers(buftuples.getIdentifier());
						tuple.fillIdentifiers();
						buftuples.add(tuple);
					}
					// can make higher buffering by modifying this test
					if (non_blocking && !buftuples.isEmpty()) {
						//org.xquark.xml.misc.Debug.print("AlgSource : buftuples = " + buftuples);
						if (resultset != null && !resultset.hasNext()) {
							XMLStatement stat = resultset.getStatement();
							resultset.close();
							resultset = null;
							stat.close();
							stat = null;
							setFinishedWhenEmpty();
						}
						break;
					}
				} else {
					XMLStatement stat = resultset.getStatement();
					resultset.close();
					resultset = null;
					stat.close();
					stat = null;
					setFinishedWhenEmpty();
					//org.xquark.xml.misc.Debug.print("AlgSource : buftuples = " + buftuples);
					return;
				}
			}
			if (resultset != null && !resultset.hasNext()) {
				XMLStatement stat = resultset.getStatement();
				resultset.close();
				resultset = null;
				stat.close();
				stat = null;
				setFinishedWhenEmpty();
			}
			//org.xquark.xml.misc.Debug.print("AlgSource : buftuples = " + buftuples);
		} catch (SAXException saxe) {
			throw new XMLDBCException("AlgSource.evaluate: " + saxe.getMessage());
		}
		if (resultset != null && !resultset.hasNext()) {
			XMLStatement stat = resultset.getStatement();
			resultset.close();
			resultset = null;
			stat.close();
			stat = null;
			setFinishedWhenEmpty();
		}
	}

	protected void finalize() {
		try {
			if (null != resultset) {
				XMLStatement stat = resultset.getStatement();
				resultset.close();
				resultset = null;
				stat.close();
				stat = null;
			}
		} catch (Exception ex) {};
	}
}
